"""Test messages.

pytest --cov-report term-missing --cov=aiounifi.messages tests/test_messages.py
"""

from unittest.mock import Mock

import pytest

from aiounifi.interfaces.messages import MessageHandler
from aiounifi.models.message import Message, MessageKey

MESSAGE_HANDLER_DATA = [
    (None, False),  # No subscriber registered
    (MessageKey.CLIENT_REMOVED, True),  # Filter correct
    (MessageKey.CLIENT, False),  # Filter incorrect
    ((MessageKey.CLIENT, MessageKey.CLIENT_REMOVED), True),  # Filter correct
]


@pytest.mark.parametrize(("message_filter", "expected"), MESSAGE_HANDLER_DATA)
async def test_message_handler(message_filter, expected):
    """Verify message handler behaves according to configured filters."""
    message_handler = MessageHandler(controller=Mock())

    filters = {}
    if message_filter:
        filters["message_filter"] = message_filter

    unsubscribe_callback = message_handler.subscribe(mock_callback := Mock(), **filters)
    assert len(message_handler) == 1
    assert unsubscribe_callback

    message_handler.handler(
        {
            "meta": {
                "rc": "ok",
                "message": MessageKey.CLIENT_REMOVED.value,
            },
            "data": [{}],
        }
    )
    assert mock_callback.called is expected

    unsubscribe_callback()
    assert len(message_handler) == 0


async def test_unsupported_message_key():
    """Validate unsupported message key handling."""
    message = Message.from_dict(
        {
            "meta": {
                "rc": "ok",
                "message": "Unsupported",
            },
            "data": [{}],
        }
    )
    assert message.meta.message == MessageKey.UNKNOWN



"""Test traffic routes API for disabling and enabling traffic routes.

pytest --cov-report term-missing --cov=aiounifi.traffic_route tests/test_traffic_routes.py
"""
from copy import deepcopy

import pytest

from aiounifi.models.traffic_route import (
    Domain,
    IPAddress,
    MatchingTarget,
    TrafficRouteSaveRequest,
)

from .fixtures import TRAFFIC_ROUTES, WIRELESS_CLIENT


@pytest.mark.parametrize("is_unifi_os", [True])
@pytest.mark.parametrize("enable", [True, False])
async def test_traffic_route_enable_request(
    mock_aioresponse, unifi_controller, unifi_called_with, enable
):
    """Test that traffic route can be enabled and disabled."""
    # Use a copy so we don't modify our fixture (this should probably be a real fixture)
    traffic_routes = deepcopy(TRAFFIC_ROUTES)

    traffic_route_enabled = traffic_routes[0]
    traffic_route_enabled_id = traffic_route_enabled["_id"]
    assert traffic_route_enabled["enabled"] is True

    traffic_route_disabled = traffic_routes[1]
    traffic_route_disabled_id = traffic_route_disabled["_id"]
    assert traffic_route_disabled["enabled"] is False

    traffic_route_id = traffic_route_disabled_id if enable else traffic_route_enabled_id
    traffic_route = traffic_route_disabled if enable else traffic_route_enabled
    # Make sure initial state is what's expected
    assert traffic_route["enabled"] is not enable

    mock_aioresponse.put(
        "https://host:8443/proxy/network/v2/api/site/default"
        + f"/trafficroutes/{traffic_route_id}",
        payload={},
    )

    await unifi_controller.request(
        TrafficRouteSaveRequest.create(traffic_route, enable)
    )

    assert unifi_called_with(
        "put",
        f"/proxy/network/v2/api/site/default/trafficroutes/{traffic_route_id}",
        json=traffic_route | {"enabled": enable},
    )


@pytest.mark.parametrize("traffic_route_payload", [TRAFFIC_ROUTES])
@pytest.mark.parametrize("is_unifi_os", [True])
@pytest.mark.parametrize("traffic_route_id", [route["_id"] for route in TRAFFIC_ROUTES])
@pytest.mark.parametrize("enable", [True, False, None])
async def test_traffic_route_save(
    mock_aioresponse,
    unifi_controller,
    unifi_called_with,
    _mock_endpoints,
    enable,
    traffic_route_id,
):
    """Test save method can enable and disable a traffic route."""
    traffic_routes = unifi_controller.traffic_routes
    await traffic_routes.update()

    traffic_route = traffic_routes[traffic_route_id]

    mock_aioresponse.put(
        "https://host:8443/proxy/network/v2/api/site/default"
        + f"/trafficroutes/{traffic_route_id}",
        payload={},
    )
    await traffic_routes.save(traffic_route, enable)
    if enable is not None:
        # If setting enabled, verify it was set
        assert unifi_called_with(
            "put",
            f"/proxy/network/v2/api/site/default/trafficroutes/{traffic_route_id}",
            json=traffic_route.raw | {"enabled": enable},
        )
    else:
        # Otherwise, verify it was not set
        assert unifi_called_with(
            "put",
            f"/proxy/network/v2/api/site/default/trafficroutes/{traffic_route_id}",
            json=traffic_route.raw,
        )


@pytest.mark.parametrize("traffic_route_payload", [TRAFFIC_ROUTES])
@pytest.mark.parametrize("is_unifi_os", [True])
@pytest.mark.parametrize("traffic_route_id", [route["_id"] for route in TRAFFIC_ROUTES])
@pytest.mark.parametrize("enable", [True, False])
async def test_traffic_route_enable_disable(
    mock_aioresponse,
    _mock_endpoints,
    unifi_called_with,
    unifi_controller,
    enable,
    traffic_route_id,
):
    """Test individual methods for enabled and disabled."""
    traffic_routes = unifi_controller.traffic_routes
    await traffic_routes.update()

    traffic_route = traffic_routes[traffic_route_id]
    traffic_route_call = traffic_routes.disable if not enable else traffic_routes.enable

    mock_aioresponse.put(
        "https://host:8443/proxy/network/v2/api/site/default"
        + f"/trafficroutes/{traffic_route_id}",
        payload={},
    )
    await traffic_route_call(traffic_routes[traffic_route_id])
    assert unifi_called_with(
        "put",
        f"/proxy/network/v2/api/site/default/trafficroutes/{traffic_route_id}",
        json=traffic_route.raw | {"enabled": enable},
    )


@pytest.mark.parametrize("is_unifi_os", [True])
async def test_no_traffic_routes(unifi_controller, _mock_endpoints, unifi_called_with):
    """Test that no traffic routes also work."""
    traffic_routes = unifi_controller.traffic_routes
    await traffic_routes.update()
    assert unifi_called_with("get", "/proxy/network/v2/api/site/default/trafficroutes")
    assert len(traffic_routes.values()) == 0


@pytest.mark.parametrize("traffic_route_payload", [TRAFFIC_ROUTES])
async def test_traffic_routes(unifi_controller, _mock_endpoints, unifi_called_with):
    """Test that we get the expected traffic route."""
    traffic_routes = unifi_controller.traffic_routes
    await traffic_routes.update()
    assert unifi_called_with("get", "/v2/api/site/default/trafficroutes")
    assert len(traffic_routes.values()) == 3

    traffic_route = traffic_routes["6468ecd4c1dd1932ad2f801c"]
    assert traffic_route.id == "6468ecd4c1dd1932ad2f801c"
    assert traffic_route.description == "Test domain rule"
    assert traffic_route.domains == [
        Domain(domain="example.com", port_ranges=[], ports=[]),
    ]
    assert traffic_route.enabled is True
    assert traffic_route.ip_addresses == []
    assert traffic_route.ip_ranges == []
    assert traffic_route.matching_target == MatchingTarget.DOMAIN
    assert traffic_route.network_id == "5a32aa4ee4b047ede36a85a8"
    assert traffic_route.next_hop == ""
    assert traffic_route.regions == []
    assert traffic_route.target_devices == [
        {"network_id": WIRELESS_CLIENT["network_id"], "type": "NETWORK"},
    ]

    traffic_route = traffic_routes["655565af1e1c2754a39388a4"]
    assert traffic_route.id == "655565af1e1c2754a39388a4"
    assert traffic_route.description == "Test all internet rule"
    assert traffic_route.domains == []
    assert traffic_route.enabled is False
    assert traffic_route.ip_addresses == []
    assert traffic_route.ip_ranges == []
    assert traffic_route.matching_target == MatchingTarget.INTERNET
    assert traffic_route.network_id == "5a32aa4ee4b047ede36a85a8"
    assert traffic_route.next_hop == ""
    assert traffic_route.regions == []
    assert traffic_route.target_devices == [
        {"network_id": WIRELESS_CLIENT["network_id"], "type": "NETWORK"},
    ]

    traffic_route = traffic_routes["655566f91e1c2754a393892c"]
    assert traffic_route.id == "655566f91e1c2754a393892c"
    assert traffic_route.description == "Test IP rule"
    assert traffic_route.domains == []
    assert traffic_route.enabled is False
    assert traffic_route.ip_addresses == [
        IPAddress(ip_or_subnet="1.1.1.1", ip_version="v4", port_ranges=[], ports=[]),
    ]
    assert traffic_route.ip_ranges == []
    assert traffic_route.matching_target == MatchingTarget.IP
    assert traffic_route.network_id == "5a32aa4ee4b047ede36a85a8"
    assert traffic_route.next_hop == ""
    assert traffic_route.regions == []
    assert traffic_route.target_devices == [
        {"network_id": WIRELESS_CLIENT["network_id"], "type": "NETWORK"},
    ]



"""WLANs as part of a UniFi network."""

from ..models.api import TypedApiResponse
from ..models.message import MessageKey
from ..models.wlan import Wlan, WlanEnableRequest, WlanListRequest, wlan_qr_code
from .api_handlers import APIHandler


class Wlans(APIHandler[Wlan]):
    """Represents WLAN configurations."""

    obj_id_key = "_id"
    item_cls = Wlan
    process_messages = (MessageKey.WLAN_CONF_UPDATED,)
    api_request = WlanListRequest.create()

    async def enable(self, wlan: Wlan) -> TypedApiResponse:
        """Block client from controller."""
        return await self.controller.request(
            WlanEnableRequest.create(wlan.id, enable=True)
        )

    async def disable(self, wlan: Wlan) -> TypedApiResponse:
        """Unblock client from controller."""
        return await self.controller.request(
            WlanEnableRequest.create(wlan.id, enable=False)
        )

    def generate_wlan_qr_code(self, wlan: Wlan) -> bytes:
        """Generate QR code based on WLAN properties."""
        return wlan_qr_code(wlan.name, wlan.x_passphrase)



"""Device outlet implementation."""


from .api import ApiItem
from .device import TypedDeviceOutletTable


class Outlet(ApiItem):
    """Represents an outlet."""

    raw: TypedDeviceOutletTable

    @property
    def name(self) -> str:
        """Name of outlet."""
        return self.raw["name"]

    @property
    def index(self) -> int:
        """Outlet index."""
        return self.raw["index"]

    @property
    def has_relay(self) -> bool | None:
        """Is the outlet controllable.

        Not reported by USP-PDU-Pro, see caps.
        """
        return self.raw.get("has_relay")

    @property
    def relay_state(self) -> bool:
        """Is outlet power on."""
        return self.raw["relay_state"]

    @property
    def cycle_enabled(self) -> bool | None:
        """Modem Power Cycle."""
        return self.raw.get("cycle_enabled")

    # Metering capabilities of outlet

    @property
    def has_metering(self) -> bool | None:
        """Is metering supported.

        Reported false by UP1 and UP6 which does not have power metering.
        Not reported by by USP-PDU-Pro, see caps.
        """
        return self.raw.get("has_metering")

    @property
    def caps(self) -> int | None:
        """Outlet capabilities.

        1: Outlet supports relay (switching)
        3: Outlet supports relay and metering
        """
        return self.raw.get("outlet_caps")

    @property
    def voltage(self) -> str | None:
        """Voltage draw of outlet."""
        return self.raw.get("outlet_voltage")

    @property
    def current(self) -> str | None:
        """Usage of outlet."""
        return self.raw.get("outlet_current")

    @property
    def power(self) -> str | None:
        """Power consumption of the outlet."""
        return self.raw.get("outlet_power")

    @property
    def power_factor(self) -> str | None:
        """Power factor."""
        return self.raw.get("outlet_power_factor")

    def __repr__(self) -> str:
        """Return the representation."""
        return f"<{self.name}: relay state {self.relay_state}>"



"""Test client configuration API.

pytest --cov-report term-missing --cov=aiounifi.clients tests/test_clients.py
"""
import pytest

from aiounifi.controller import Controller


@pytest.mark.parametrize(
    "site_payload",
    [
        [
            {
                "_id": "5e231c10931eb902acf25112",
                "name": "default",
                "desc": "Default",
                "attr_hidden_id": "default",
                "attr_no_delete": True,
                "role": "admin",
            }
        ]
    ],
)
async def test_sites(unifi_controller: Controller, _mock_endpoints: None) -> None:
    """Test sites class."""
    sites = unifi_controller.sites
    await sites.update()
    assert len(sites.items()) == 1

    site = sites["5e231c10931eb902acf25112"]
    assert site.site_id == "5e231c10931eb902acf25112"
    assert site.description == "Default"
    assert site.hidden_id == "default"
    assert site.name == "default"
    assert site.no_delete is True
    assert site.role == "admin"



"""Traffic routes as part of a UniFi network."""
from dataclasses import dataclass
from enum import StrEnum
from typing import NotRequired, Self, TypedDict

from .api import ApiItem, ApiRequestV2


class MatchingTarget(StrEnum):
    """Possible matching targets for a traffic rule."""

    DOMAIN = "DOMAIN"
    IP = "IP"
    INTERNET = "INTERNET"
    REGION = "REGION"


class PortRange(TypedDict):
    """Port range type definition."""

    port_start: int
    port_stop: int


class IPAddress(TypedDict):
    """IP Address for which traffic route is applicable type definition."""

    ip_or_subnet: str
    ip_version: str
    port_ranges: list[PortRange]
    ports: list[int]


class IPRange(TypedDict):
    """IP Range type definition."""

    ip_start: str
    ip_stop: str
    ip_version: str


class TargetDevice(TypedDict):
    """Target device to which the traffic route applies."""

    client_mac: NotRequired[str]
    network_id: NotRequired[str]
    type: str


class Domain(TypedDict):
    """A target domain for a traffic route."""

    domain: str
    port_ranges: list[PortRange]
    ports: list[int]


class TypedTrafficRoute(TypedDict):
    """Traffic route type definition."""

    _id: str
    description: str
    domains: list[Domain]
    enabled: bool
    ip_addresses: list[IPAddress]
    ip_ranges: list[IPRange]
    matching_target: MatchingTarget
    network_id: str
    next_hop: str
    regions: list[str]
    target_devices: list[TargetDevice]


@dataclass
class TrafficRouteListRequest(ApiRequestV2):
    """Request object for traffic route list."""

    @classmethod
    def create(cls) -> Self:
        """Create traffic route request."""
        return cls(method="get", path="/trafficroutes", data=None)


@dataclass
class TrafficRouteSaveRequest(ApiRequestV2):
    """Request object for saving a traffic route.

    To modify a route, you must make sure the `raw` attribute of the TypedTrafficRoute is modified.
    The properties provide convient access for reading, however do not provide means of setting values.
    """

    @classmethod
    def create(
        cls, traffic_route: TypedTrafficRoute, enable: bool | None = None
    ) -> Self:
        """Create traffic route save request."""
        if enable is not None:
            traffic_route["enabled"] = enable
        return cls(
            method="put",
            path=f"/trafficroutes/{traffic_route['_id']}",
            data=traffic_route,
        )


class TrafficRoute(ApiItem):
    """Represent a traffic route configuration."""

    raw: TypedTrafficRoute

    @property
    def id(self) -> str:
        """ID of traffic route."""
        return self.raw["_id"]

    @property
    def description(self) -> str:
        """Description given by user to traffic route."""
        return self.raw["description"]

    @property
    def domains(self) -> list[Domain]:
        """What IP addresses are matched against by this traffic route.

        Note: Domain matching requires the client devices to use the UniFi gateway as the DNS server.
        """
        return self.raw["domains"]

    @property
    def enabled(self) -> bool:
        """Is traffic route enabled."""
        return self.raw["enabled"]

    @property
    def ip_addresses(self) -> list[IPAddress]:
        """What IP addresses are matched against by this traffic route."""
        return self.raw["ip_addresses"]

    @property
    def ip_ranges(self) -> list[IPRange]:
        """What IP addresses are matched against by this traffic route."""
        return self.raw["ip_ranges"]

    @property
    def matching_target(self) -> MatchingTarget:
        """What target is matched by this traffic route."""
        return self.raw["matching_target"]

    @property
    def network_id(self) -> str:
        """Network ID that this rule applies to."""
        return self.raw["network_id"]

    @property
    def next_hop(self) -> str:
        """Used for defining a static route."""
        return self.raw["next_hop"]

    @property
    def regions(self) -> list[str]:
        """Regions that the rule applies to."""
        return self.raw["regions"]

    @property
    def target_devices(self) -> list[TargetDevice]:
        """What target devices are affected by this traffic route."""
        return self.raw["target_devices"]



"""Test system information API."""
import pytest

from aiounifi.controller import Controller
from aiounifi.models.system_information import SystemInformationRequest

from .fixtures import SYSTEM_INFORMATION


async def test_sys_info_request(mock_aioresponse, unifi_controller, unifi_called_with):
    """Test power cycle port work."""
    mock_aioresponse.get("https://host:8443/api/s/default/stat/sysinfo", payload={})
    await unifi_controller.request(SystemInformationRequest.create())
    assert unifi_called_with("get", "/api/s/default/stat/sysinfo")


@pytest.mark.parametrize("system_information_payload", [[SYSTEM_INFORMATION]])
async def test_system_information(
    unifi_controller: Controller, _mock_endpoints: None
) -> None:
    """Test port forwarding interface and model."""
    system_information = unifi_controller.system_information
    await system_information.update()
    assert len(system_information.values()) == 1

    sys_info = next(iter(system_information.values()))
    assert sys_info.anonymous_controller_id == "24f81231-a456-4c32-abcd-f5612345385f"
    assert sys_info.device_type == "UDMPRO"
    assert sys_info.hostname == "UDMP"
    assert sys_info.ip_address == ["1.2.3.4"]
    assert sys_info.is_cloud_console is False
    assert sys_info.name == "UDMP"
    assert sys_info.previous_version == "7.4.156"
    assert sys_info.update_available is False
    assert sys_info.uptime == 1196290
    assert sys_info.version == "7.4.162"



"""Clients are devices on a UniFi network."""

from dataclasses import dataclass
from typing import Self, TypedDict, cast

from .api import ApiItem, ApiRequest


class TypedClient(TypedDict):
    """Client type definition."""

    _id: str
    _is_guest_by_uap: bool
    _is_guest_by_ugw: bool
    _is_guest_by_usw: bool
    _last_seen_by_uap: int
    _last_seen_by_ugw: int
    _last_seen_by_usw: int
    _uptime_by_uap: int
    _uptime_by_ugw: int
    _uptime_by_usw: int
    anomalies: int
    ap_mac: str
    assoc_time: int
    authorized: bool
    blocked: bool
    bssid: str
    bytes_r: int
    ccq: int
    channel: int
    dev_cat: int
    dev_family: int
    dev_id: int
    dev_id_override: int
    dev_vendor: int
    device_name: str
    dhcpend_time: int
    disconnect_timestamp: int
    eagerly_discovered: bool
    essid: str
    fingerprint_engine_version: str
    fingerprint_override: bool
    fingerprint_source: int
    first_seen: int
    fixed_ip: str
    fw_version: str
    gw_mac: str
    hostname: str
    hostname_source: str
    idletime: int
    ip: str
    is_11r: bool
    is_guest: bool
    is_wired: bool
    last_seen: bool
    latest_assoc_time: bool
    mac: str
    name: str
    network: str
    network_id: str
    noise: int
    noted: bool
    os_class: int
    os_name: int
    oui: str
    powersave_enabled: bool
    qos_policy_applied: bool
    radio: str
    radio_name: str
    radio_proto: str
    rssi: int
    rx_bytes: int
    rx_packets: int
    rx_rate: int
    satisfaction: int
    score: int
    signal: int
    site_id: str
    sw_depth: int
    sw_mac: str
    sw_port: int
    tx_bytes: int
    tx_packets: int
    tx_power: int
    tx_rate: int
    tx_retries: int
    uptime: int
    use_fixedip: bool
    user_id: str
    usergroup_id: str
    vlan: int
    wifi_tx_attempts: int
    wired_rate_mbps: int
    wired_tx_bytes: int
    wired_rx_bytes: int
    wired_tx_packets: int
    wired_rx_packets: int
    wired_tx_bytes_r: int
    wired_rx_bytes_r: int


@dataclass
class AllClientListRequest(ApiRequest):
    """Request object for all clients list."""

    @classmethod
    def create(cls) -> Self:
        """Create all clients list request."""
        return cls(method="get", path="/rest/user")


@dataclass
class ClientListRequest(ApiRequest):
    """Request object for active client list."""

    @classmethod
    def create(cls) -> Self:
        """Create active client list request."""
        return cls(method="get", path="/stat/sta")


@dataclass
class ClientBlockRequest(ApiRequest):
    """Request object for client block."""

    @classmethod
    def create(cls, mac: str, block: bool) -> Self:
        """Create client block request."""
        return cls(
            method="post",
            path="/cmd/stamgr",
            data={
                "cmd": "block-sta" if block else "unblock-sta",
                "mac": mac,
            },
        )


@dataclass
class ClientReconnectRequest(ApiRequest):
    """Request object for client reconnect."""

    @classmethod
    def create(cls, mac: str) -> Self:
        """Create client reconnect request."""
        return cls(
            method="post",
            path="/cmd/stamgr",
            data={
                "cmd": "kick-sta",
                "mac": mac,
            },
        )


@dataclass
class ClientRemoveRequest(ApiRequest):
    """Request object for client removal."""

    @classmethod
    def create(cls, macs: list[str]) -> Self:
        """Create client removal request."""
        return cls(
            method="post",
            path="/cmd/stamgr",
            data={
                "cmd": "forget-sta",
                "macs": macs,
            },
        )


class Client(ApiItem):
    """Represents a client network device."""

    raw: TypedClient

    @property
    def access_point_mac(self) -> str:
        """MAC address of access point."""
        return self.raw.get("ap_mac", "")

    @property
    def association_time(self) -> int:
        """When was client associated with controller."""
        return self.raw.get("assoc_time", 0)

    @property
    def blocked(self) -> bool:
        """Is client blocked."""
        return self.raw.get("blocked", False)

    @property
    def device_name(self) -> str:
        """Device name of client."""
        return self.raw.get("device_name", "")

    @property
    def essid(self) -> str:
        """ESSID client is connected to."""
        return self.raw.get("essid", "")

    @property
    def firmware_version(self) -> str:
        """Firmware version of client."""
        return self.raw.get("fw_version", "")

    @property
    def first_seen(self) -> int:
        """When was client first seen."""
        return self.raw.get("first_seen", 0)

    @property
    def fixed_ip(self) -> str:
        """List IP if fixed IP is configured."""
        return self.raw.get("fixed_ip", "")

    @property
    def hostname(self) -> str:
        """Hostname of client."""
        return self.raw.get("hostname", "")

    @property
    def idle_time(self) -> int:
        """Idle time of client."""
        return self.raw.get("idletime", 0)

    @property
    def ip(self) -> str:
        """IP of client."""
        return self.raw.get("ip", "")

    @property
    def is_guest(self) -> bool:
        """Is client guest."""
        return self.raw.get("is_guest", False)

    @property
    def is_wired(self) -> bool:
        """Is client wired."""
        return self.raw.get("is_wired", False)

    @property
    def last_seen(self) -> int:
        """When was client last seen."""
        return self.raw.get("last_seen", 0)

    @property
    def last_seen_by_access_point(self) -> int:
        """When was client last seen by access point."""
        return self.raw.get("_last_seen_by_uap", 0)

    @property
    def last_seen_by_gateway(self) -> int:
        """When was client last seen by gateway."""
        return self.raw.get("_last_seen_by_ugw", 0)

    @property
    def last_seen_by_switch(self) -> int:
        """When was client last seen by network switch."""
        return self.raw.get("_last_seen_by_usw", 0)

    @property
    def latest_association_time(self) -> int:
        """When was client last associated with controller."""
        return self.raw.get("latest_assoc_time", 0)

    @property
    def mac(self) -> str:
        """MAC address of client."""
        return self.raw.get("mac", "")

    @property
    def name(self) -> str:
        """Name of client."""
        return self.raw.get("name", "")

    @property
    def oui(self) -> str:
        """Vendor string for client MAC."""
        return self.raw.get("oui", "")

    @property
    def powersave_enabled(self) -> bool | None:
        """Powersave functionality enabled for wireless client."""
        return self.raw.get("powersave_enabled")

    @property
    def site_id(self) -> str:
        """Site ID client belongs to."""
        return self.raw.get("site_id", "")

    @property
    def switch_depth(self) -> int | None:
        """How many layers of switches client is in."""
        return self.raw.get("sw_depth")

    @property
    def switch_mac(self) -> str:
        """MAC for switch client is connected to."""
        return self.raw.get("sw_mac", "")

    @property
    def switch_port(self) -> int | None:
        """Switch port client is connected to."""
        return self.raw.get("sw_port")

    @property
    def rx_bytes(self) -> int:
        """Bytes received over wireless connection."""
        return self.raw.get("rx_bytes", 0)

    @property
    def rx_bytes_r(self) -> float:
        """Bytes recently received over wireless connection."""
        return cast(float, self.raw.get("rx_bytes-r", 0.0))

    @property
    def tx_bytes(self) -> int:
        """Bytes transferred over wireless connection."""
        return self.raw.get("tx_bytes", 0)

    @property
    def tx_bytes_r(self) -> float:
        """Bytes recently transferred over wireless connection."""
        return cast(float, self.raw.get("tx_bytes-r", 0.0))

    @property
    def uptime(self) -> int:
        """Uptime of client."""
        return self.raw.get("uptime", 0)

    @property
    def uptime_by_access_point(self) -> int:
        """Uptime of client observed by access point."""
        return self.raw.get("_uptime_by_uap", 0)

    @property
    def uptime_by_gateway(self) -> int:
        """Uptime of client observed by gateway."""
        return self.raw.get("_uptime_by_ugw", 0)

    @property
    def uptime_by_switch(self) -> int:
        """Uptime of client observed by network switch."""
        return self.raw.get("_uptime_by_usw", 0)

    @property
    def wired_rate_mbps(self) -> int:
        """Wired rate in Mbps."""
        return self.raw.get("wired_rate_mbps", 0)

    @property
    def wired_rx_bytes(self) -> int:
        """Bytes received over wired connection."""
        return cast(int, self.raw.get("wired-rx_bytes", 0))

    @property
    def wired_rx_bytes_r(self) -> float:
        """Bytes recently received over wired connection."""
        return cast(float, self.raw.get("wired-rx_bytes-r", 0.0))

    @property
    def wired_tx_bytes(self) -> int:
        """Bytes transferred over wired connection."""
        return cast(int, self.raw.get("wired-tx_bytes", 0))

    @property
    def wired_tx_bytes_r(self) -> float:
        """Bytes recently transferred over wired connection."""
        return cast(float, self.raw.get("wired-tx_bytes-r", 0.0))



"""Port forwarding in a UniFi network."""
from dataclasses import dataclass
from typing import NotRequired, Self, TypedDict

from .api import ApiItem, ApiRequest


class TypedPortForward(TypedDict):
    """Port forward type definition."""

    _id: str
    dst_port: str
    enabled: NotRequired[bool]
    fwd_port: str
    fwd: str
    name: str
    pfwd_interface: str
    proto: str
    site_id: str
    src: str


@dataclass
class PortForwardListRequest(ApiRequest):
    """Request object for port forward list."""

    @classmethod
    def create(cls) -> Self:
        """Create port forward list request."""
        return cls(method="get", path="/rest/portforward")


@dataclass
class PortForwardEnableRequest(ApiRequest):
    """Request object for enabling port forward."""

    @classmethod
    def create(cls, port_forward: "PortForward", enable: bool) -> Self:
        """Create enable port forward request."""
        data = port_forward.raw.copy()
        data["enabled"] = enable
        return cls(
            method="put",
            path=f"/rest/portforward/{data['_id']}",
            data=data,
        )


class PortForward(ApiItem):
    """Represents a port forward configuration."""

    raw: TypedPortForward

    @property
    def id(self) -> str:
        """Unique ID of port forward."""
        return self.raw["_id"]

    @property
    def destination_port(self) -> str:
        """Destination port."""
        return self.raw["dst_port"]

    @property
    def enabled(self) -> bool:
        """Is port forward enabled."""
        return self.raw.get("enabled", False)

    @property
    def forward_port(self) -> str:
        """Forwarded port."""
        return self.raw["fwd_port"]

    @property
    def forward_ip(self) -> str:
        """IP address to forward."""
        return self.raw["fwd"]

    @property
    def name(self) -> str:
        """Name of port forward."""
        return self.raw["name"]

    @property
    def port_forward_interface(self) -> str:
        """Interface to expose port forward."""
        return self.raw["pfwd_interface"]

    @property
    def protocol(self) -> str:
        """Protocol to forward.

        tcp_udp
        """
        return self.raw["proto"]

    @property
    def site_id(self) -> str:
        """Site id port forward belongs to."""
        return self.raw["site_id"]

    @property
    def source(self) -> str:
        """Source."""
        return self.raw["src"]



"""DPI Restrictions as part of a UniFi network."""


from ..models.dpi_restriction_group import (
    DPIRestrictionGroup,
    DpiRestrictionGroupListRequest,
)
from ..models.message import MessageKey
from .api_handlers import APIHandler


class DPIRestrictionGroups(APIHandler[DPIRestrictionGroup]):
    """Represents DPI Group configurations."""

    obj_id_key = "_id"
    item_cls = DPIRestrictionGroup
    process_messages = (MessageKey.DPI_GROUP_ADDED, MessageKey.DPI_GROUP_UPDATED)
    remove_messages = (MessageKey.DPI_GROUP_REMOVED,)
    api_request = DpiRestrictionGroupListRequest.create()



"""Test controller.

pytest --cov-report term-missing --cov=aiounifi.controller tests/test_controller.py
"""

import ssl

from aiohttp import ClientSession, client_exceptions, web
import pytest
import trustme

from aiounifi import (
    AiounifiException,
    BadGateway,
    Forbidden,
    LoginRequired,
    NoPermission,
    RequestError,
    ResponseError,
    ServiceUnavailable,
    TwoFaTokenRequired,
    Unauthorized,
)
from aiounifi.controller import Controller
from aiounifi.models.api import ApiRequest, ApiRequestV2
from aiounifi.models.configuration import Configuration

from .fixtures import LOGIN_UNIFIOS_JSON_RESPONSE, SITE_RESPONSE

EMPTY_RESPONSE = {"meta": {"rc": "ok"}, "data": []}


@pytest.mark.parametrize("is_unifi_os", [True, False])
async def test_check_unifi(
    mock_aioresponse, unifi_controller, unifi_called_with, is_unifi_os
):
    """Test validating if controller is hosted on UniFi OS."""
    if is_unifi_os:
        mock_aioresponse.get(
            "https://host:8443",
            content_type="text/html",
            headers={"x-csrf-token": "012"},
        )
    else:
        mock_aioresponse.get(
            "https://host:8443",
            content_type="application/octet-stream",
            status=302,
        )
    await unifi_controller.connectivity.check_unifi_os()
    assert unifi_controller.connectivity.is_unifi_os is is_unifi_os
    assert unifi_called_with("get", "", allow_redirects=False)


@pytest.mark.parametrize("is_unifi_os", [True, False])
async def test_login(
    mock_aioresponse, unifi_controller, unifi_called_with, is_unifi_os
):
    """Test logging in to controller."""
    if is_unifi_os:
        mock_aioresponse.get(
            "https://host:8443",
            content_type="text/html",
            headers={"x-csrf-token": "012"},
        )
        mock_aioresponse.post(
            "https://host:8443/api/auth/login",
            payload=LOGIN_UNIFIOS_JSON_RESPONSE,
            headers={"x-csrf-token": "123"},
            content_type="text/json",
        )
        await unifi_controller.connectivity.login()
        assert unifi_called_with(
            "post",
            "/api/auth/login",
            json={"username": "user", "password": "pass", "remember": True},
        )
    else:
        mock_aioresponse.post(
            "https://host:8443/api/login",
            payload={"meta": {"rc": "ok"}, "data": []},
            content_type="application/json",
        )
        await unifi_controller.connectivity.login()
        assert unifi_called_with(
            "post",
            "/api/login",
            json={"username": "user", "password": "pass", "remember": True},
        )


@pytest.mark.parametrize("is_unifi_os", [True, False])
async def test_controller_login(
    mock_aioresponse, unifi_controller, unifi_called_with, is_unifi_os
):
    """Test logging in to controller."""
    if is_unifi_os:
        mock_aioresponse.get(
            "https://host:8443",
            content_type="text/html",
            headers={"x-csrf-token": "012"},
        )
        mock_aioresponse.post(
            "https://host:8443/api/auth/login",
            payload=LOGIN_UNIFIOS_JSON_RESPONSE,
            headers={"x-csrf-token": "123"},
            content_type="text/json",
        )
        await unifi_controller.login()
        assert unifi_called_with(
            "post",
            "/api/auth/login",
            json={"username": "user", "password": "pass", "remember": True},
        )
    else:
        mock_aioresponse.get(
            "https://host:8443", content_type="application/octet-stream", status=302
        )
        mock_aioresponse.post(
            "https://host:8443/api/login",
            payload={"meta": {"rc": "ok"}, "data": []},
        )
        await unifi_controller.login()
        assert unifi_called_with(
            "post",
            "/api/login",
            json={"username": "user", "password": "pass", "remember": True},
        )
    assert unifi_called_with("get", "", allow_redirects=False)


async def test_relogin_success(mock_aioresponse, unifi_controller):
    """Test controller communicating with a UniFi OS controller with retries."""
    mock_aioresponse.get(
        "https://host:8443",
        body="<html>",
        headers={"x-csrf-token": "012"},
        content_type="text/html",
        status=200,
    )

    await unifi_controller.connectivity.check_unifi_os()
    assert unifi_controller.connectivity.is_unifi_os

    mock_aioresponse.post(
        "https://host:8443/api/auth/login",
        payload=LOGIN_UNIFIOS_JSON_RESPONSE,
        content_type="text/json",
        headers={"x-csrf-token": "123"},
        status=200,
    )

    await unifi_controller.connectivity.login()

    mock_aioresponse.get(
        "https://host:8443/proxy/network/api/s/default/stat/sta",
        payload=EMPTY_RESPONSE,
        content_type="text/json",
        status=200,
    )
    await unifi_controller.clients.update()

    # After a login failure we retry once
    mock_aioresponse.get(
        "https://host:8443/proxy/network/api/s/default/stat/device",
        body="<html>AUTH FAILED</html>",
        content_type="text/html",
        status=401,
    )

    mock_aioresponse.get(
        "https://host:8443",
        body="<html>",
        headers={"x-csrf-token": "012"},
        content_type="text/html",
        status=200,
    )
    mock_aioresponse.post(
        "https://host:8443/api/auth/login",
        payload=LOGIN_UNIFIOS_JSON_RESPONSE,
        headers={"x-csrf-token": "563"},
        content_type="text/json",
        status=200,
    )
    mock_aioresponse.get(
        "https://host:8443/proxy/network/api/s/default/stat/device",
        payload=EMPTY_RESPONSE,
        content_type="text/json",
        status=200,
    )

    await unifi_controller.devices.update()


async def test_relogin_fails(mock_aioresponse, unifi_controller):
    """Test controller communicating with a UniFi OS controller with retries."""
    mock_aioresponse.get(
        "https://host:8443",
        body="<html>",
        headers={"x-csrf-token": "012"},
        content_type="text/html",
        status=200,
    )

    await unifi_controller.connectivity.check_unifi_os()
    assert unifi_controller.connectivity.is_unifi_os
    assert len(mock_aioresponse.requests) == 1

    mock_aioresponse.post(
        "https://host:8443/api/auth/login",
        payload=LOGIN_UNIFIOS_JSON_RESPONSE,
        headers={"x-csrf-token": "123"},
        content_type="text/json",
        status=200,
    )

    await unifi_controller.connectivity.login()

    mock_aioresponse.get(
        "https://host:8443/proxy/network/api/s/default/stat/sta",
        payload=EMPTY_RESPONSE,
        content_type="text/json",
        status=200,
    )
    await unifi_controller.clients.update()

    # After a login failure we retry once
    mock_aioresponse.get(
        "https://host:8443/proxy/network/api/s/default/stat/device",
        body="<html>AUTH FAILED</html>",
        content_type="text/html",
        status=401,
    )
    mock_aioresponse.post(
        "https://host:8443/api/auth/login",
        payload=LOGIN_UNIFIOS_JSON_RESPONSE,
        headers={"x-csrf-token": "456"},
        content_type="text/json",
        status=401,
    )

    with pytest.raises(LoginRequired):
        await unifi_controller.devices.update()

    # After a login failure and retry, we do
    # not retry over and over
    mock_aioresponse.get(
        "https://host:8443/proxy/network/api/s/default/stat/device",
        body="<html>AUTH FAILED</html>",
        content_type="text/html",
        status=401,
    )
    with pytest.raises(LoginRequired):
        await unifi_controller.devices.update()


@pytest.mark.parametrize("site_payload", [SITE_RESPONSE["data"]])
async def test_controller(
    unifi_controller, unifi_called_with, _mock_endpoints, _new_ws_data_fn
):
    """Test controller communicating with a non UniFiOS UniFi controller."""
    await unifi_controller.initialize()

    assert unifi_called_with("get", "/api/s/default/stat/sta")
    assert unifi_called_with("get", "/api/s/default/rest/user")
    assert unifi_called_with("get", "/api/s/default/stat/device")
    assert unifi_called_with("get", "/api/s/default/rest/dpiapp")
    assert unifi_called_with("get", "/api/s/default/rest/dpigroup")
    assert unifi_called_with("get", "/api/s/default/rest/portforward")
    assert unifi_called_with("get", "/api/self/sites")
    assert unifi_called_with("get", "/api/s/default/stat/sysinfo")
    assert unifi_called_with("get", "/v2/api/site/default/trafficroutes")
    assert unifi_called_with("get", "/v2/api/site/default/trafficrules")
    assert unifi_called_with("get", "/api/s/default/rest/wlanconf")

    assert len(unifi_controller.clients.items()) == 0
    assert len(unifi_controller.clients_all.items()) == 0
    assert len(unifi_controller.devices.items()) == 0
    assert len(unifi_controller.outlets.items()) == 0
    assert len(unifi_controller.ports.items()) == 0
    assert len(unifi_controller.dpi_apps.items()) == 0
    assert len(unifi_controller.dpi_groups.items()) == 0
    assert len(unifi_controller.port_forwarding.items()) == 0
    assert len(unifi_controller.sites.items()) == 1
    assert len(unifi_controller.system_information.items()) == 0
    assert len(unifi_controller.traffic_routes.items()) == 0
    assert len(unifi_controller.traffic_rules.items()) == 0
    assert len(unifi_controller.wlans.items()) == 0


@pytest.mark.parametrize(("is_unifi_os", "site_payload"), [(True, SITE_RESPONSE)])
async def test_unifios_controller(
    mock_aioresponse,
    unifi_controller,
    unifi_called_with,
    _mock_endpoints,
    _new_ws_data_fn,
):
    """Test controller communicating with a UniFi OS controller."""
    mock_aioresponse.post(
        "https://host:8443/api/auth/login",
        payload=LOGIN_UNIFIOS_JSON_RESPONSE,
        headers={"x-csrf-token": "123"},
        content_type="text/json",
    )
    await unifi_controller.connectivity.login()
    await unifi_controller.initialize()

    assert unifi_called_with(
        "get",
        "/proxy/network/api/s/default/stat/sta",
        headers={"x-csrf-token": "123"},
    )
    assert unifi_called_with(
        "get",
        "/proxy/network/api/s/default/stat/device",
        headers={"x-csrf-token": "123"},
    )
    assert unifi_called_with(
        "get",
        "/proxy/network/api/s/default/rest/user",
        headers={"x-csrf-token": "123"},
    )
    assert unifi_called_with(
        "get",
        "/proxy/network/api/self/sites",
        headers={"x-csrf-token": "123"},
    )
    assert unifi_called_with(
        "get",
        "/proxy/network/v2/api/site/default/trafficroutes",
        headers={"x-csrf-token": "123"},
    )
    assert unifi_called_with(
        "get",
        "/proxy/network/v2/api/site/default/trafficrules",
        headers={"x-csrf-token": "123"},
    )
    assert unifi_called_with(
        "get",
        "/proxy/network/api/s/default/rest/wlanconf",
        headers={"x-csrf-token": "123"},
    )


async def test_unifios_controller_no_csrf_token(
    mock_aioresponse, unifi_controller, unifi_called_with
):
    """Test controller communicating with a UniFi OS controller without csrf token."""
    mock_aioresponse.get(
        "https://host:8443",
        content_type="text/html",
    )
    await unifi_controller.connectivity.check_unifi_os()
    assert unifi_controller.connectivity.is_unifi_os
    assert unifi_called_with(
        "get",
        "",
        allow_redirects=False,
    )

    mock_aioresponse.post(
        "https://host:8443/api/auth/login",
        payload=LOGIN_UNIFIOS_JSON_RESPONSE,
        content_type="text/json",
    )
    await unifi_controller.connectivity.login()
    assert unifi_called_with(
        "post",
        "/api/auth/login",
        json={"username": "user", "password": "pass", "remember": True},
    )


test_data = [
    ({"status": 401}, LoginRequired),
    ({"status": 403}, Forbidden),
    ({"status": 404}, ResponseError),
    ({"status": 502}, BadGateway),
    ({"status": 503}, ServiceUnavailable),
    ({"exception": client_exceptions.ClientError}, RequestError),
    (
        {"payload": {"meta": {"msg": "api.err.LoginRequired", "rc": "error"}}},
        LoginRequired,
    ),
    (
        {"payload": {"meta": {"msg": "api.err.Invalid", "rc": "error"}}},
        Unauthorized,
    ),
    (
        {"payload": {"meta": {"msg": "api.err.NoPermission", "rc": "error"}}},
        NoPermission,
    ),
    (
        {"payload": {"meta": {"msg": "api.err.Ubic2faTokenRequired", "rc": "error"}}},
        TwoFaTokenRequired,
    ),
]


@pytest.mark.parametrize(("unwanted_behavior", "expected_exception"), test_data)
async def test_controller_raise_expected_exception(
    mock_aioresponse, unifi_controller, unwanted_behavior, expected_exception
):
    """Verify request raise login required on a 401."""
    mock_aioresponse.post("https://host:8443/api/login", **unwanted_behavior)
    with pytest.raises(expected_exception):
        await unifi_controller.connectivity.login()


@pytest.mark.parametrize("traffic_rule_status", [404])
async def test_initialize_handles_404(unifi_controller, _mock_endpoints):
    """Validate initialize does not abort on exception."""
    await unifi_controller.initialize()


api_request_data = [
    (
        ApiRequest,
        "/api/s/default",
        {"payload": {"meta": {"msg": "api.err.LoginRequired", "rc": "error"}}},
        LoginRequired,
    ),
    (
        ApiRequest,
        "/api/s/default",
        {"payload": {"meta": {"msg": "api.err.Invalid", "rc": "error"}}},
        Unauthorized,
    ),
    (
        ApiRequest,
        "/api/s/default",
        {"payload": {"meta": {"msg": "api.err.NoPermission", "rc": "error"}}},
        NoPermission,
    ),
    (
        ApiRequest,
        "/api/s/default",
        {"payload": {"meta": {"msg": "api.err.Ubic2faTokenRequired", "rc": "error"}}},
        TwoFaTokenRequired,
    ),
    (
        ApiRequest,
        "/api/s/default",
        {"payload": {"meta": {"msg": "api.err.OtherError", "rc": "error"}}},
        AiounifiException,
    ),
    (
        ApiRequestV2,
        "/v2/api/site/default",
        {"payload": {"errorCode": 1, "message": "api.err.LoginRequired"}},
        LoginRequired,
    ),
    (
        ApiRequestV2,
        "/v2/api/site/default",
        {"payload": {"errorCode": 2, "message": "api.err.Invalid"}},
        Unauthorized,
    ),
    (
        ApiRequestV2,
        "/v2/api/site/default",
        {"payload": {"errorCode": 3, "message": "api.err.NoPermission"}},
        NoPermission,
    ),
    (
        ApiRequestV2,
        "/v2/api/site/default",
        {"payload": {"errorCode": 4, "message": "api.err.Ubic2faTokenRequired"}},
        TwoFaTokenRequired,
    ),
    (
        ApiRequestV2,
        "/v2/api/site/default",
        {"payload": {"errorCode": 5, "message": "api.err.OtherError"}},
        AiounifiException,
    ),
]


@pytest.mark.parametrize(("api_request", "path", "input", "expected"), api_request_data)
async def test_api_request_error_handling(
    mock_aioresponse,
    unifi_controller: Controller,
    api_request,
    path,
    input,
    expected,
):
    """Verify request raise login required on a 401."""
    mock_aioresponse.get(f"https://host:8443{path}/test", **input)
    with pytest.raises(expected):
        await unifi_controller.connectivity.request(api_request("get", "/test"))


@pytest.mark.parametrize(("unwanted_behavior", "expected_exception"), test_data)
async def test_api_request_generic_error_handling(
    mock_aioresponse,
    unifi_controller: Controller,
    unwanted_behavior,
    expected_exception,
):
    """Verify request raise login required on a 401."""
    mock_aioresponse.get("https://host:8443/api/s/default/test", **unwanted_behavior)
    with pytest.raises(expected_exception):
        await unifi_controller.connectivity.request(ApiRequest("get", "/test"))


@pytest.mark.parametrize(
    "unsupported_message", ["device:update", "unifi-device:sync", "unsupported"]
)
async def test_handle_unsupported_events(
    unifi_controller, unsupported_message, _new_ws_data_fn
):
    """Test controller properly ignores unsupported events."""
    unifi_controller.ws_state_callback.reset_mock()
    _new_ws_data_fn({"meta": {"message": unsupported_message}})
    unifi_controller.ws_state_callback.assert_not_called()

    assert len(unifi_controller.clients.items()) == 0


async def test_websocket(aiohttp_server) -> None:
    """Test positive websocket."""

    tls_certificate_authority = trustme.CA()
    tls_certificate = tls_certificate_authority.issue_server_cert(
        "localhost", "127.0.0.1", "::1"
    )
    ssl_context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
    tls_certificate.configure_cert(ssl_context)

    async def handler(request):
        ws = web.WebSocketResponse()
        await ws.prepare(request)

        await ws.send_json({"meta": {"message": "device:update"}})
        await ws.send_json(
            {
                "meta": {"rc": "ok", "message": "dpigroup:add"},
                "data": [
                    {
                        "name": "dpi group",
                        "site_id": "5f3edd27ba4cc806a19f2d9c",
                        "_id": "61783dbdc1773a18c0c61ef6",
                    }
                ],
            }
        )

        await ws.close()
        return ws

    app = web.Application()
    app.router.add_get("/wss/s/default/events", handler)
    await aiohttp_server(app, port=8443, ssl=ssl_context)

    config = Configuration(
        ClientSession(), "0.0.0.0", username="user", password="pass", ssl_context=False
    )
    controller = Controller(config)
    await controller.start_websocket()

    assert len(controller.dpi_groups.items()) == 1
